# Copyright 2012-2014 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
# http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.

import time
import logging
import datetime
import hashlib

import dateutil.parser
from dateutil.tz import tzutc

import botocore
from botocore.utils import (
    CachedProperty,
    datetime2timestamp
)
from botocore.compat import (
        total_seconds
)


logger = logging.getLogger(__name__)

class SSOTokenFetcher(object):
    # The device flow RFC defines the slow down delay to be an additional
    # 5 seconds:
    # https://tools.ietf.org/html/draft-ietf-oauth-device-flow-15#section-3.5
    _SLOW_DOWN_DELAY = 5
    # The default interval of 5 is also defined in the RFC (see above link)
    _DEFAULT_INTERVAL = 5
    _EXPIRY_WINDOW = 0 * 60
    _CLIENT_REGISTRATION_TYPE = 'public'
    _GRANT_TYPE = 'urn:ietf:params:oauth:grant-type:device_code'

    def __init__(
            self, sso_region, client_creator, cache=None,
            on_pending_authorization=None, time_fetcher=None, sleep=None,
    ):
        self._sso_region = sso_region
        self._client_creator = client_creator
        self._on_pending_authorization = on_pending_authorization

        if time_fetcher is None:
            time_fetcher = self._utc_now
        self._time_fetcher = time_fetcher

        if sleep is None:
            sleep = time.sleep
        self._sleep = sleep

        if cache is None:
            cache = {}
        self._cache = cache

    def _utc_now(self):
        return datetime.datetime.now(tzutc())

    def _parse_if_needed(self, value):
        if isinstance(value, datetime.datetime):
            return value
        return dateutil.parser.parse(value)

    def _is_expired(self, response):
        end_time = self._parse_if_needed(response['expiresAt'])
        seconds = total_seconds(end_time - self._time_fetcher())
        return seconds < self._EXPIRY_WINDOW

    @CachedProperty
    def _client(self):
        config = botocore.config.Config(
            region_name=self._sso_region,
            signature_version=botocore.UNSIGNED,
        )
        return self._client_creator('sso-oidc', config=config)

    def _register_client(self):
        timestamp = datetime2timestamp(self._time_fetcher())
        response = self._client.register_client(
            # NOTE: As far as I know client name will never be returned to the
            # user. For now we'll just use botocore-client with the timestamp.
            clientName='botocore-client-%s' % int(timestamp),
            clientType=self._CLIENT_REGISTRATION_TYPE,
        )
        expires_at = response['clientSecretExpiresAt']
        expires_at = datetime.datetime.fromtimestamp(expires_at, tzutc())
        registration = {
            'clientId': response['clientId'],
            'clientSecret': response['clientSecret'],
            'expiresAt': expires_at,
        }
        return registration

    def _registration(self):
        # Registration with the OIDC endpoint is regional, is good for roughly
        # 90 days, and can be shared. This is currently scoped to individual
        # tools and as such each tool has their own cached client id.
        cache_key = 'botocore-client-id-%s' % self._sso_region
        if cache_key in self._cache:
            registration = self._cache[cache_key]
            if not self._is_expired(registration):
                return registration

        registration = self._register_client()
        self._cache[cache_key] = registration
        return registration

    def _authorize_client(self, start_url, registration):
        # NOTE: The authorization response is not cached. These responses are
        # short lived (currently only 10 minutes) and can only be exchanged for
        # a token once. Having multiple clients share this is problematic.
        response = self._client.start_device_authorization(
            clientId=registration['clientId'],
            clientSecret=registration['clientSecret'],
            startUrl=start_url,
        )
        expires_in = datetime.timedelta(seconds=response['expiresIn'])
        authorization = {
            'deviceCode': response['deviceCode'],
            'userCode': response['userCode'],
            'verificationUri': response['verificationUri'],
            'verificationUriComplete': response['verificationUriComplete'],
            'expiresAt': self._time_fetcher() + expires_in,
        }
        if 'interval' in response:
            authorization['interval'] = response['interval']
        return authorization

    def _poll_for_token(self, start_url):
        registration = self._registration()
        authorization = self._authorize_client(start_url, registration)

        if self._on_pending_authorization:
            # This callback can display the user code / verification URI
            # so the user knows the page to go to. Potentially, this call
            # back could even be used to auto open a browser.
            self._on_pending_authorization(**authorization)

        interval = authorization.get('interval', self._DEFAULT_INTERVAL)
        # NOTE: This loop currently relies on the service to either return
        # a valid token or a ExpiredTokenException to break the loop. If this
        # proves to be problematic it may be worth adding an additional
        # mechanism to control timing this loop out.
        while True:
            try:
                response = self._client.create_token(
                    grantType=self._GRANT_TYPE,
                    clientId=registration['clientId'],
                    clientSecret=registration['clientSecret'],
                    deviceCode=authorization['deviceCode'],
                )
                expires_in = datetime.timedelta(seconds=response['expiresIn'])
                token = {
                    'startUrl': start_url,
                    'region': self._sso_region,
                    'accessToken': response['accessToken'],
                    'expiresAt': self._time_fetcher() + expires_in
                }
                return token
            except self._client.exceptions.SlowDownException:
                interval += self._SLOW_DOWN_DELAY
            except self._client.exceptions.AuthorizationPendingException:
                pass
            except self._client.exceptions.ExpiredTokenException:
                logger.error("The pending authorization to retrieve an SSO token has expired. The device authorization flow to retrieve an SSO token must be restarted.")
            self._sleep(interval)

    def _token(self, start_url, force_refresh):
        cache_key = hashlib.sha1(start_url.encode('utf-8')).hexdigest()
        # Only obey the token cache if we are not forcing a refresh.
        if not force_refresh and cache_key in self._cache:
            token = self._cache[cache_key]
            if not self._is_expired(token):
                return token

        token = self._poll_for_token(start_url)
        self._cache[cache_key] = token
        return token

    def fetch_token(self, start_url, force_refresh=False):
        return self._token(start_url, force_refresh)
    
    def store_token(self, start_url, create_token_response):
        response = create_token_response
        expires_in = datetime.timedelta(seconds=response['expiresIn'])
        token = {
            'startUrl': start_url,
            'region': self._sso_region,
            'accessToken': response['accessToken'],
            'expiresAt': self._time_fetcher() + expires_in
        }

        cache_key = hashlib.sha1(start_url.encode('utf-8')).hexdigest()
        self._cache[cache_key] = token

    def is_sso_token_valid(self, start_url):
        cache_key = hashlib.sha1(start_url.encode('utf-8')).hexdigest()
        
        is_valid = False
        
        if cache_key in self._cache:
            token = self._cache[cache_key]
            if not self._is_expired(token):
                is_valid = True
        
        return is_valid



import datetime
import json
import os


SSO_TOKEN_DIR = os.path.expanduser(
    os.path.join('~', '.aws', 'sso', 'cache')
)

def _serialize_utc_timestamp(obj):
    if isinstance(obj, datetime.datetime):
        return obj.strftime('%Y-%m-%dT%H:%M:%SZ')
    return obj


def sso_json_dumps(obj):
    return json.dumps(obj, default=_serialize_utc_timestamp)


from datetime import datetime, timezone
from itertools import chain
from logging import getLogger
import time
import boto3
from .utils import SSO_TOKEN_DIR, sso_json_dumps
from .vendored.botocore.utils import SSOTokenFetcher
import botocore.session
from botocore.credentials import JSONFileCache
from botocore.utils import SSOTokenLoader
from botocore.exceptions import SSOTokenLoadError

logger = getLogger(__name__)

bc_session = botocore.session.Session()

class AWSSSO:
    def __init__(self, start_url:str, sso_region:str):
        self._start_url = start_url
        self._sso_region = sso_region

        self._token_fetcher = SSOTokenFetcher(
            sso_region=sso_region,
            client_creator=bc_session.create_client,
            cache=JSONFileCache(SSO_TOKEN_DIR, dumps_func=sso_json_dumps)
        )
        registration = self._token_fetcher._registration()
        self._client_id = registration['clientId']
        self._client_secret = registration['clientSecret']

        boto3_session = boto3.session.Session(region_name=sso_region)

        self.sso_oidc_client = boto3_session.client('sso-oidc')
        self.sso_client = boto3_session.client('sso')

    def _get_cache_key(self):
        """Generate cache key including access token to auto-invalidate on token change"""
        access_token = self._get_sso_access_token()
        # If no token, use a fixed key (cache will be empty anyway)
        token_part = hash(access_token) if access_token else 'no-token'
        return hash(f'{self._start_url}::{self._sso_region}::{token_part}')

    
    def start_device_authorization(self):
        return self.sso_oidc_client.start_device_authorization(
            clientId = self._client_id,
            clientSecret = self._client_secret,
            startUrl=self._start_url
        )

    def create_token(self, device_code, store_token=True):
        retry = True
        while retry:
            retry = False
            try:
                create_token_response = self.sso_oidc_client.create_token(
                    clientId = self._client_id,
                    clientSecret = self._client_secret,
                    grantType='urn:ietf:params:oauth:grant-type:device_code',
                    deviceCode=device_code
                )

                if store_token:
                    self._token_fetcher.store_token(self._start_url, create_token_response)
                    
                return {"status":"successful"}
            except self.sso_oidc_client.exceptions.SlowDownException:
                time.sleep(5)
                retry = True
            except self.sso_oidc_client.exceptions.AuthorizationPendingException:
                return {"status":"pending"}
            except self.sso_oidc_client.exceptions.ExpiredTokenException as e:
                return {"status":"error","error": "Token expired. Please restart the authorization process."}
            
        return {"status":"error","error": "Unknown error occurred during token creation."}
    
    def get_role_credentials(self, account_id:str, role_name:str):
        access_token = self._get_sso_access_token()
        return self.sso_client.get_role_credentials(
            accessToken=access_token,
            accountId=account_id,
            roleName=role_name
        )
    
    def has_valid_access_token(self):
        access_token = self._get_sso_access_token()
        return access_token is not None

    def _get_sso_access_token(self):
        token = None
        token_loader = SSOTokenLoader(JSONFileCache(SSO_TOKEN_DIR, dumps_func=sso_json_dumps))

        try:
            token = token_loader(self._start_url)
        except SSOTokenLoadError:
            logger.debug("Token not found")
        except Exception as e:
            logger.debug(e)

        if token is None:
            return None

        expires_at = token.get('expiresAt')

        if expires_at and datetime.strptime(expires_at, '%Y-%m-%dT%H:%M:%S%z').replace(tzinfo=timezone.utc) > datetime.now(tz=timezone.utc):
            return token.get('accessToken')
        else:
            return None
    
    def get_aws_accounts(self):
        cache_key = self._get_cache_key()

        global aws_accounts_cache
        
        if cache_key in aws_accounts_cache:
            return aws_accounts_cache[cache_key]

        access_token = self._get_sso_access_token()
        accounts = list(chain.from_iterable(
                page['accountList'] 
                for page in self.sso_client.get_paginator('list_accounts').paginate(accessToken=access_token)))
        aws_accounts_cache[cache_key] = accounts
        return accounts
    
    def get_aws_account_roles(self, account_id:str):
        cache_key = f'{self._get_cache_key()}::roles::{account_id}'

        global aws_account_roles_cache
        
        if cache_key in aws_account_roles_cache:
            return aws_account_roles_cache[cache_key]

        access_token = self._get_sso_access_token()
        roles = list(chain.from_iterable(
            page['roleList'] 
            for page in self.sso_client.get_paginator('list_account_roles').paginate(accessToken=access_token, accountId=account_id))
        )
        aws_account_roles_cache[cache_key] = roles
        return roles
    
    def get_account_id_by_account_name(self, account_name:str):
        accounts = self.get_aws_accounts()
        for account in accounts:
            if account.get('accountName') == account_name:
                return account.get('accountId')
        
        logger.debug(f'Account name {account_name} not found')
        return None
    
    def get_boto3_session(self, account_id:str, sso_role_name:str, assumed_role_arn:str = None):
        if assumed_role_arn is not None:
            return self.get_boto3_session_using_assumed_role_credentials(account_id, sso_role_name, assumed_role_arn)
        else:
            return self.get_boto3_session_using_sso_role_credentials(account_id, sso_role_name)

    def get_boto3_session_using_sso_role_credentials(self, account_id:str, sso_role_name:str):
        cache_key = f'{self._get_cache_key()}::sso-role::{account_id}::{sso_role_name}'

        global _credentials
        global _sessions

        if cache_key in _credentials and cache_key in _sessions:
            if _credentials.get(cache_key).get('expiration') > int(datetime.now(tz=timezone.utc).timestamp() * 1000):
                return _sessions.get(cache_key)

        response = self.get_role_credentials(account_id, sso_role_name)

        role_credentials = response.get('roleCredentials', {})

        _credentials[cache_key] = role_credentials
        _sessions[cache_key] = boto3.session.Session(
            aws_access_key_id=role_credentials["accessKeyId"],
            aws_secret_access_key=role_credentials["secretAccessKey"],
            aws_session_token=role_credentials["sessionToken"]
        )

        return _sessions[cache_key]

    def get_boto3_session_using_assumed_role_credentials(self, account_id:str, sso_role_name: str, assumed_role_arn: str):
        cache_key = f'{self._get_cache_key()}::assumed-role::{account_id}::{sso_role_name}::{assumed_role_arn}'

        global _credentials
        global _sessions

        if cache_key in _credentials and cache_key in _sessions:
            if _credentials.get(cache_key).get('Expiration') > datetime.now(tz=timezone.utc):
                return _sessions.get(cache_key)

        def _assumed_role():
            sso_role_session = self.get_boto3_session_using_sso_role_credentials(account_id, sso_role_name)
            sts_client = sso_role_session.client('sts', region_name=self._sso_region)
            return sts_client.assume_role(
                RoleArn=assumed_role_arn,
                RoleSessionName=f'vwsre-{datetime.now().timestamp()}'
            )
        
        response = _assumed_role()

        _credentials[cache_key] = response.get('Credentials')
        _sessions[cache_key] = boto3.session.Session(
            aws_access_key_id=_credentials[cache_key]['AccessKeyId'],
            aws_secret_access_key=_credentials[cache_key]['SecretAccessKey'],
            aws_session_token=_credentials[cache_key]['SessionToken']
        )
        return _sessions[cache_key]

aws_accounts_cache = {}
aws_account_roles_cache = {}
_credentials = {}
_sessions = {}